;; Generated package description from .el  -*- flymake-clippyno-byte-compile: t -*-
(define-package "flymake-clippy" "1.1.0" "Flymake backend for Clippy" '((emacs "27")) :commit "38aeda6a8f4090b00f3c7fdb5188f80e04c556de" :authors '(("Michael Kirkland" . "mak.kirkland@proton.me")) :maintainer '("Michael Kirkland" . "mak.kirkland@proton.me") :keywords '("languages" "tools") :url "https://github.com/mak-kirkland/flymake-clippy")
